/***********************************************************************
*$Header:   J:/aim/gedcom/gedlib/vcs/gedask.c_v   1.8   21 Nov 1995 07:53:38   ODJENSEN  $
*
*$config$="/K! /L/* /R* /Mgedask.c"
*
*   FILE NAME: GEDASK.C           
*
*   DESCRIPTION:
*       This file contains functions which find context within a record
*
*   ROUTINES:
*       NODE * ged_ask(NODE * obj, byte * msg, int nth)
*       NODE * ged_ask_partial(NODE * obj, byte * msg, int nth)
*       NODE * ged_hunt_tags(NODE * obj, byte * msg, int *nth)
*       int    ged_inherits(NODE *node1, NODE *node2)
*       NODE * ged_ask_xref(NODE *obj, byte *xref);
*       int    ged_match_xrefs(int unreferencedLevel, NODE * obj,
*                              void *xref, void *unreferencedArg2)
*
*   MODIFICATION HISTORY:
*   _________________________________________________________________________
*    |  Date  | Who |  Comments
*    -------------------------------------------------------------------------
*$Log:   J:/aim/gedcom/gedlib/vcs/gedask.c_v  $
 * 
 *    Rev 1.8   21 Nov 1995 07:53:38   ODJENSEN
 * Changed prototypes to new style. Removed assignment within conditional
 * warnings.
 * 
 *    Rev 1.7   03 Jun 1993 08:51:50   fhdodj
 * Replaced the old gedask functions with the corresponding functions from 
 * gedaskst.c.  Also changed the arguments for ged_inherits to be NODE *
 * and byte *.
 * 
 *    Rev 1.7   03 Jun 1993 08:38:42   fhdodj
 * Replaced the old gedask functions with the corresponding functions from
 * gedaskst.c
 * 
 *    Rev 1.6   25 Mar 1992 15:44:36   fhdodj
 * Added function ged_ask_partial().
 * 
 *    Rev 1.5   21 Oct 1991 08:55:48   fhdodj
 * changed char to byte.
 * 
 *    Rev 1.4   28 Jun 1991 11:38:20   fhdkrf
 * Added keywords for PolyDoc
 * 
 *    Rev 1.3   16 May 1991 13:37:40   odj
 * Moved ged_ask_string to its own file (gedaskst.c), and moved ged_match 
 * functions to the file gedmisc.c.
 * 
 *    Rev 1.2   03 May 1991 11:44:00   odj
 * Added check in ged_ask_string.  If ged_make_path fails, return NULL.
 * 
 *    Rev 1.1   15 Feb 1991 10:06:46   odj
 * moved void * into ifdefs for non ansi compilers who use char * instead.
 * 
 *    Rev 1.0   20 Dec 1990 08:16:32   odj
 * Initial revision.
***********************************************************************/

#include "gedcom.h"

/********************************************************************
*
*        NAME: ged_ask
*
*        DESCRIPTION:
*            Gets the nth occurance of the node in obj at the end of
*            the path msg. In the future, if there is no value, an
*            attempt will be made to compute a value.
*
*        RETURN VALUE:
*            Returns the nth occurance of the node at the end of the
*            path or NULL if unsuccesful. If a value less than 0 is
*            given for nth, the first value will be returned. If a
*            value greater than the number of occurances is given,
*            NULL will be returned.
*
*        CALLS: ged_get_child, ged_hunt_tags
*
********************************************************************/
NODE * ged_ask(NODE *obj,  /* The tree in which we will follow path msg */
               byte *msg,  /* The path to follow */
               int   nth)  /* Which occurance to return */
    {

        if ( !msg && (nth < 2) ) /* If no msg and we are looking for the */
            return(obj);         /* first occurance we are already there */
        if ( !msg || !*msg )  /* If no msg and we are not looking for the */
            return(NULL);     /* next occurance, there isn't one */
        obj = ged_get_child(obj);
        if ( !obj )  /* If no tree to look in, we must fail. */
            return(NULL);
        return(ged_hunt_tags(obj, msg, &nth));
    }

/********************************************************************
*
*        NAME: ged_ask_partial
*
*        DESCRIPTION:
*            Gets the nth occurance of the node in obj at the end of
*            the partial path msg. In the future, if there is no value, an
*            attempt will be made to compute a value.
*
*        RETURN VALUE:
*            Returns the nth occurance of the node at the end of the
*            path or NULL if unsuccesful. If a value less than 0 is
*            given for nth, the first value will be returned. If a
*            value greater than the number of occurances is given,
*            NULL will be returned.
*
*        CALLS: ged_get_child, ged_hunt_tags, ged_get_next_node
*
********************************************************************/
NODE * ged_ask_partial(NODE *obj,  /* The tree in which to follow path msg */
                       byte *msg,  /* The path to follow */
                       int   nth)  /* Which occurance to return */
    {
    NODE *ret_val;
    NODE *tmp_obj;
    NODE *start_obj = obj;

        if ( !msg && (nth < 2) ) /* If no msg and we are looking for the */
            return(obj);         /* first occurance we are already there */
        if ( !msg || !*msg )  /* If no msg and we are not looking for the */
            return(NULL);     /* next occurance, there isn't one */
        if ( nth <= 0 )
	    nth = 1;
	while(obj && (nth > 0))
	    {
	    tmp_obj = ged_get_child(obj);
	    if (tmp_obj)
                ret_val = ged_hunt_tags(tmp_obj, msg, &nth);
            if ((obj = ged_get_next_node(obj)) == start_obj)
                {
	        obj = NULL;
		ret_val = NULL;
		}
	    }
        return(ret_val);
    }

/********************************************************************
*
*        NAME: ged_hunt_tags
*
*        DESCRIPTION:
*            Traverses the path in msg to find the nth occurance of
*            the node in the tree obj at the end of the path.
*
*        RETURN VALUE:
*            The nth occurance of the node at the end of the path
*            msg in obj, or NULL if not found. If a value less than
*            one is given for nth, the first occurance will be
*            returned. If a value for nth is given that is greater
*            than the number of occurances, NULL will be returned.
*
*        CALLS: ged_match, ged_skip_word, ged_prev_word, ged_get_tag,
*               ged_get_parent
*
*        CALLED BY: ged_ask
*
********************************************************************/
NODE * ged_hunt_tags(NODE *obj,  /* The tree to look in */
                     byte *msg,  /* The path describing what to look for */
                     int  *nth)  /* Which occurance to return */
    {
    NODE *lastObj = obj->parent;  /* contains the last obj position */
    NODE *start = obj->parent;    /* tells where we started. */
    byte *msgStart = msg;
    byte *tmpPos;

        while (isspace(*msg))
            msg++;
        while (msg)
            {

            /* find the next part of the path                   */
            while (obj && ged_inherits(obj, msg))
                obj = obj->sibling;

            /* try again when no more obj or need next answer.  */
            if (!obj || (!ged_skip_word(msg) && (--(*nth) > 0)))
                {
                if (!obj)
                    {
                    obj = lastObj;
                    msg = ged_prev_word(msg, msgStart);
                    }
                while (!obj->sibling && msg)
                    {
                    obj = obj->parent;
                    msg = ged_prev_word(msg, msgStart);
                    }
                obj = obj->sibling;
                lastObj = ged_get_parent(obj);
                }

            /* if there is more in msg keep looking.            */
            else if ((tmpPos = ged_skip_word(msg)) != 0)
                {
                msg = tmpPos;
                lastObj = obj;
                obj = obj->child;
                }

            /* The right answer was found */
            else if (*nth <= 0)
                return(obj);
            }
        if (start->sibling == obj)
            return(NULL);
        return(obj);
    }

/********************************************************************
*
*        NAME: ged_inherits
*
*        DESCRIPTION:
*            Will determine if node1 is equivalent to node2.
*            Currently only checks that tag1 equals tag2. In the
*            future will check for things like marriage is
*            equivalent to banns.
*
*        RETURN VALUE:
*            returns 1 if false, and 0 if true.
*
*        CALLS: ged_match_tags
*
*        CALLED BY: ged_hunt_tags
*
********************************************************************/
int ged_inherits(NODE *node, byte *tag)
    {
	return(ged_match(ged_get_tag(node), tag));
    }

/********************************************************************
*
*        NAME: ged_ask_xref
*
*        DESCRIPTION:
*            Finds the node in obj at the current level, or at a
*            lower level with the given Xref value.
*
*        RETURN VALUE:
*            Returns the node with the given Xref value, or NULL if
*            it cannot be found.
*
*        CALLS: ged_apply_tree, ged_match_xrefs
*
********************************************************************/
NODE * ged_ask_xref(NODE *obj,   /* The tree in which to find xref */
                    byte *xref)  /* The cross reference to find */
    {

        if (!obj || !xref)  /* If no tree, or no xref to look for, fail */
            return(NULL);
        return(ged_apply_tree(obj, 0, 1, ged_match_xrefs, xref, NULL));
    }

/********************************************************************
*
*        NAME: ged_match_xrefs
*
*        DESCRIPTION:
*            Compares two Xrefs.
*
*        RETURN VALUE:
*            Returns 0 if equal, or a non-zero value if not.
*
*        CALLS: ged_get_xref, ged_match
*
*        CALLED BY: ged_hunt_xref, ged_apply_tree
*
********************************************************************/
int ged_match_xrefs(int unreferencedLevel, NODE * obj, void * xref,
                    void * unreferencedArg)
    {
        return(ged_match(ged_get_xref(obj), (byte *)xref));
    }
