 /****************************************************************************
 *$Header:   J:/gedcom/gedhash/vcs/gedhash.c_v   1.2   16 Sep 1994 16:08:28   ODJENSEN  $
 *
 *      Copyright (c) 1991  Corporation of the President of
 *                          the CHURCH of JESUS CHRIST of LATTER-DAY SAINTS.
 *
 *      All rights reserved.  Unauthorized reproduction of this software is
 *      prohibited and is in violaton of United States copyright laws.
 *
 ****************************************************************************/
/*******************************************************************
 *
 * FILE: GEDHASH.C
 *
 * MODIFICATION HISTORY:
 * ----------------------------------------------------------------------
 * |  Date           | Who | Comments
 * ----------------------------------------------------------------------
 * | Dec 90          | LAM | Code Written by Larry A. Maddocks
 * ----------------------------------------------------------------------
 * | 4/19/91         | LAM | Had trouble with memory hash. Disk works, though.
 * ----------------------------------------------------------------------
 * | 08/14/91 01:45pm| LAM | Fixed memory-based hashing. Church mice had eaten some code.
 * ----------------------------------------------------------------------
 * | 08/14/91 01:46pm| LAM | Compiled with Borland C
 * ----------------------------------------------------------------------
 * | 10/02/91 09:45am| LAM | Took out key & val parameters from gedhash_copy()
 * ----------------------------------------------------------------------
 *$Log:   J:/gedcom/gedhash/vcs/gedhash.c_v  $
 * 
 *    Rev 1.2   16 Sep 1994 16:08:28   ODJENSEN
 * Fixed case where table is put on disk and pool is destoryed, but pool
 * pointer is not set to NULL.
 * 
 *    Rev 1.1   03 Aug 1994 11:30:52   odjensen
 ********************************************************************/

#include "gedcom.h"
#include "GEDHSTAT.H"

#ifdef TC          /* if running Turbo or Borland C */
#define _splitpath fnsplit  /* Used in gedh_floppy_prompt() */
#define _makepath fnmerge
#endif

/* set by floppy_prompt() & returned by gedhash_get_file() */
static char NEW_FILE_NAME[MIN_FNAME_SIZ];
static char DRIVE[3]; /* set by floppy_prompt() & returned by gedhash_get_drive() */


/************************************************************************
 * NAME: gedhash_create_table
 *
 * DESCRIPTION: Allocates memory for and initializes a hash table.
 *
 * RETURN VALUE: Returns a pointer to the hash table, or NULL if unsuccessful.
 *
 * PARAMETERS:
 *
 *
 * htab:  input:     Pointer to HASH_TAB struct.  Must have been allocated.
 *       output:     The complete struct is initialized.
 *
 * perm_file: input: Name of file that will store the hash table.  If this is
 *                   an existing hash file, the values in the file will be used
 *                   during program execution, else a new permanent file will be
 *                   created.  If the argument is NULL, the hash file will be
 *                   memory-based until max_memory is exceeded, at which time hashing
 *                   operations will be performed on a temprary disk-based file.
 *
 *
 *          output: None
 *
 *
 * ged_file: input: File that this hash table is recieving its data from, or NULL
 *                  if not applicable.
 *          output: None.
 *
 *
 * hash_size: input: Size of hash table.
 *
 * max_memory input: Maximum bytes of memory a memory-based hash table should use.
 * 
 ************************************************************************/
HASH_TAB *gedhash_create_table(HASH_TAB *htab, char *perm_file, HSIZ hash_size, MEMSIZ max_memory)
{
  POOL *oldPool = ged_get_pool(); /* remember what the old pool is */
  unsigned pool_size = /* Initial pool size max is 65k, but more mem can be allocated */
      (max_memory > (MEMSIZ)65000) ? (unsigned)65000 : (unsigned int)max_memory;

  hash_size = gedh_prime(hash_size); /* Make sure this is a prime number */

  /* Make sure that BLKLEN is big enough to handle at least a HASH_TAB & KVDISK struck */
  if (!(BLKLEN >  sizeof(KVDISK)) )
    {
    return((HASH_TAB *)NULL);
    }

  *(htab->gedfn) = '\0'; /* Initialize gedcom data file */
  htab->key_count = 0L; /* Initialize number of keys in hash table to 0. */
  htab->h = hash_size;  /* Size of hash table */

  /* If a file name was passed in then allocate space for it. If not, allocate
   * space for a temporary file name: */
  if (perm_file && *perm_file) /* if a file name was passed in */
    {
    HASH_TAB *dtab;  /* disk hash table */
    #ifdef CYBER
    HASH_TAB *new_dtab; /* Used to word-align */
    #endif
        
    char buffer[sizeof(HASH_TAB) + 15]; /* buffer to read in header (15 isnot needed)*/
    
    strcpy(htab->hfn, perm_file);
    htab->isMemory = FALSE; /* File is not going to memory */
    if ((htab->fh = open(htab->hfn,O_RDWR|O_BINARY|O_CREAT,S_IWRITE|S_IREAD)) == -1)
      ged_error(OPEN_ERR, NULL);
    
    /******** CHANGE FOLLOWING when header is removed from hash table */
    if (filelength(htab->fh) < sizeof(HASH_TAB) + 15) /* If file does not already exist */
      gedh_diskSetup(htab);
    else
      {
      GED_EREAD(htab->fh, buffer, 15); /* read special header block into buf */
      if (memcmp(buffer, "GED_HASH_TAB 1", 14)) /* If don't compare */
        gedh_diskSetup(htab); /* copy hash file to disk */
      else
        {
        GED_EREAD(htab->fh, buffer, sizeof(HASH_TAB) ); /* read special header block into buf */
        
        dtab = (HASH_TAB *)(buffer);
        #ifdef CYBER /* word-align */   
        new_dtab = (HASH_TAB *)gedh_word_align(dtab);
        *new_dtab = *dtab; /* copy to new dtab */
        dtab = new_dtab;
        #endif
        /* dtab contains HASH_TAB struct stored in perm. hash file.
        *  Now copy some fields to htab */
        htab->h = dtab->h; /* get size of hash from file */
        htab->it_pos = dtab->it_pos;
        htab->disk_iterate = dtab->disk_iterate;
        
        /* if we didn't pass in an associated gedcom file but there is one stored in dtab */
        if ( (!htab->gedfn || !*(htab->gedfn) ) && dtab->gedfn)
          strcpy(htab->gedfn, dtab->gedfn);
        htab->ged_ext = dtab->ged_ext;
        htab->key_count = dtab->key_count;
        }
      }
    }
  else /* if user passed in no file name */
    {
    *(htab->hfn) = '\0';  /* Set hash file name to NULL */
    htab->fh = 0; /* No hash file handle */
    htab->isMemory = TRUE; /* Is this going to memory?  Yes. */
    htab->maxMemory = htab->memoryLeft = max_memory; /* don't need if using disk */
    htab->memoryLeft -= (MEMSIZ)((htab->h) * sizeof(KVPAIR *));
    if (htab->memoryLeft < (MEMSIZ)1)
      gedh_diskSetup(htab); /* setup on disk */
    else
      {  
      htab->pool = ged_set_pool(ged_create_pool(pool_size, (unsigned)0));
      htab->table=(KVPAIR **)ALLOCMEM(htab, (LEN)(htab->h) * sizeof(KVPAIR *));
  
      /* Memory-based should never need h greater than size_t */
      memset( htab->table, '\0', ((size_t)htab->h * sizeof(KVPAIR *)) );
      ged_set_pool(oldPool); /* reset to old pool */
      }
    }
  return(htab);
  }

/****************************************************************************
 *
 * NAME: gedhash_get_file
 *
 * DESCRIPTION: floppy_prompt() sends a code 412 to ged_error().  The user-
 * defined error-handler should use gedhash_get_file() and gedhash_get_drive() when
 * prompting the user to change floppies. See floppy_prompt()
 *
 * RETURN:  The file name on the floppy that should be inserted.
 *
 ****************************************************************************/
char *gedhash_get_file(void)
   {
   return NEW_FILE_NAME;
   } 

/****************************************************************************
 *
 * NAME: gedhash_get_drive
 *
 * DESCRIPTION: floppy_prompt() sends a code 412 to ged_error().  The user-
 * defined error-handler should use gedhash_get_file() and gedhash_get_drive() when
 * prompting the user to change floppies. See floppy_prompt()
 * 
 * RETURN:  The drive letter that the user should insert the floppy disk
 *
 ****************************************************************************/
char *gedhash_get_drive(void) 
  { 
  return DRIVE;
  }

/****************************************************************************
 *
 * NAME: gedhash_fopen
 *
 * DESCRIPTION: Opens the gedcom file that is associated with htab and copies
 * gedcom_path to htab->gedfn.  htab->ged_ext is initialized to -1. This
 * function is not called by gedhash_create_table() because not always do we
 * have a gedcom file associated with a hash table.
 * 
 * RETURN:  A pointer to the opened file if successful, else NULL.
 * 
 * CALLED BY: User, gedhash_create_xref()
 * 
 * PARAMETERS:
 *
 * htab:  input: Pointer to HASH_TAB struct.
 *       output: htab->gedfn gets path & name of file, htab->fg get file pointer
 *
 * gedcom_path:  Input: path & name of gedcom file
 *       output: None
 * 
 * mode: Input: Specifies the type of access requested for the file.
 *
 ****************************************************************************/
FILE *gedhash_fopen(HASH_TAB *htab, char *gedcom_path, char *mode)
  {
  if (!htab || !gedcom_path || !*gedcom_path || !mode || !*mode)
    {
    ged_error(GEDFN_ERR, NULL);
    return (FILE *)NULL;
    }

  /* -1 means that gedcom file extension is ".GED".  If we have gedcom that spans
   * multiple disks, this will be changed to 0 for ".G00", 1 for ".G01", etc. */
  htab->ged_ext = -1;

  strcpy(htab->gedfn, gedcom_path);
  return (htab->fg = ged_efopen(gedcom_path, mode));
  }

/****************************************************************************
 *
 * NAME: gedhash_copy
 *
 * DESCRIPTION:  Copies contents from_table to to_table.  to_table does not
 * need to be empty.  Both must have been previously created using
 * gedhash_create.  Neither their respective sizes nor their being on disk or
 * in memory affects the operation.
 *
 * RETURN: SUCCESS(which is equal to 1) if Successful, NO_SUCCESS( = 0) otherwise.
 *
 * PARAMETERS:
 *
 * htab_dest:  input: Pointer to destination hash table.
 *            output: All the key/val pairs from htab_source are inserted.
 * 
 * htab_source input: Pointer to Source hash table.
 *            output: None
 *
 ****************************************************************************/
int gedhash_copy(HASH_TAB *htab_dest, HASH_TAB *htab_source)
   {
   KSIZ keyLen;
   LEN  valLen;
   char key[BLKLEN], val[BLKLEN];

   if (!htab_source || !htab_dest)
     return NO_SUCCESS;
   gedhash_iterate_reset(htab_source);
   while (gedhash_iterate(htab_source, key, &keyLen, val, &valLen)==SUCCESS)
     {
     if (gedhash_insert(htab_dest, key, keyLen, val, valLen) != SUCCESS)
       return NO_SUCCESS;
     }
   return SUCCESS;
   }

/****************************************************************************
 *
 * NAME: gedhash_iterate
 *
 * DESCRIPTION: Successively returns every key/value pair in the hash table.
 * The order of returning pairs is undefined. gedhash_iterate_reset() must be
 * called before calling this function.  The function is called repeatedly
 * in a loop until it returns zero, indicating that all pairs have been returned.
 * Increments htab->mem_iterate and htab->it_pos.
 *
 * RETURN:  Non-zero if a key/value pair is returned, else zero, indicating that
 * all pairs have been returned.
 *
 * PARAMETERS:
 *
 * htab:  input: Pointer to HASH_TAB struct.
 *       output: Increments htab->mem_iterate and htab->it_pos.
 * 
 * key:   input: Storage location for key
 *       output: The key in the next record.
 * 
 * keyLen: output: Length of key
 *
 * val:   input: Storage location for val
 *       output: The val in the next record.
 *
 * valLen: output: Length of val
 *
 ****************************************************************************/
int gedhash_iterate(HASH_TAB *htab, void *key, KSIZ *keyLen, void *val, LEN *valLen)
  {
  if (!htab->fh)  /* If not on disk */
    {
    KVPAIR *kv = htab->mem_iterate;  /* points to current position in collision list in memory*/
    ITSIZ *it_pos = &(htab->it_pos); /* Points to currant row, or beginning of collision list */

    for (; *it_pos < htab->h; (*it_pos)++, kv = htab->table[*it_pos] )
      {
        if (kv) /* If there is a key/val in this collision list */
        {
        *valLen = kv->valLen;
        *keyLen = kv->keyLen;
        memcpy((char *)val, kv->val, *valLen);
        memcpy((char *)key, kv->key, *keyLen);
        htab->mem_iterate = kv ? kv->next : kv;
        return SUCCESS;
        }
      }

    return NO_SUCCESS; /* Nothing to return */
    }
  return gedh_iterate_disk(htab, (char *)key, keyLen, (char *)val, valLen);
  }

/****************************************************************************
 * NAME: gedhash_iterate_reset
 *
 * DESCRIPTION: Resets internal hash_tab state so that successive calls to
 * gedhash_interate() will return every key/value pair in the table.
 *
 * PARAMETERS:
 *
 * htab:  input: Pointer to HASH_TAB struct.
 *       output: htab->it_pos == 0; htab->iterate_kv points to the first value in
 *               the hash table.
 ****************************************************************************/
void gedhash_iterate_reset(HASH_TAB *htab)
  {
  htab->it_pos = (ITSIZ)0;
  if (!htab->fh)  /* If not on disk */
    htab->mem_iterate = htab->table[htab->it_pos];
  else
    htab->disk_iterate = 0L; /* Seek position on disk */
  }

/*************************************************************************
 *
 * NAME:  gedh_floppy_prompt
 *
 * DESCRIPTION:  This function is used for gedcom files that span multiple
 * files.  The SEEK struct is stored in the value fields in the hash table.
 * This contains the lseek location of a xref.  It also contains a ged_ext
 * field, which contains -1 if the gedcom file extention is .ged.  It 
 * contains 0 if the gedcom file extension is .g00, 1 if the extension is .g01,
 * etc.
 * 
 * PARAMETERS:
 * 
 * htab:    input: Pointer to HASH_TAB struct.
 *          output: htab->fg (file pointer) and htab->gedfn and htab->ged_ext.
 * 
 * ged_ext: input: Shows what file we must be in.  If this doesn't match
 *                 htab->ged_ext, and the right file cannot be found, then
 *                 user prompted to change disks.
 *          output: None.
 *
 *************************************************************************/
void gedh_floppy_prompt(HASH_TAB *htab, int ged_ext)
#ifdef CYBER  /* This has not been, & may not need to be, ported to CYBER */
  {
  }
#else
  {
  char dir[PATH_SIZ];        /* _splitpath arguments */
  char fname[MIN_FNAME_SIZ]; /*      "     "    " */
  char ext[EXT_SIZ];         /*      "     "    " */                          
                                                                                                

  if (ged_ext != htab->ged_ext)
    {
    fclose(htab->fg);  /* Close gedcom data file */
    _splitpath(htab->gedfn, DRIVE, dir, fname, ext); /* split htab->gedfn into fields. */
      
    if (ged_ext == -1) /* if -1 then we are looking for a .ged file from user */
      {
      sprintf(ext,
        #ifdef TC
        ".GED"
        #else
        "GED"
        #endif
        ); /* Create next file that user must insert */
      }
    else
      {
      sprintf(ext, "%sG%02d",
        #ifdef TC /* If compiling with Turbo C */
        "."
        #else
        ""
        #endif
        , ged_ext
        ); /* Create next file that user must insert */
      }
    sprintf(NEW_FILE_NAME, "%s.%s", fname, ext); /* Create next file for user to insert */
    _makepath( htab->gedfn, DRIVE, dir, fname, ext );
    while ((htab->fg = ged_efopen(htab->gedfn, "rb")) == (FILE *)NULL)
      {
      ged_error(CHNG_FLOPPY, NULL); /* User must insert a floppy with the right .G00 */
      }
    htab->ged_ext = ged_ext;
    }
  }
#endif



/*************************************************************************
 *
 * NAME:  gedhash_find_xref
 *
 * DESCRIPTION:  Returns the record being pointed to by the xref in the
 *               value field of dataCtxt.
 * 
 * RETURN: A NODE pointer to the record, if found, else NULL.
 *
 * PARAMETERS:
 *   htab:      input: Hash table
 * 
 *   dataCtxt:  input: The value field in this gedcom line contains the
 *              xref of the record that will be returned
 *              ouput: None.
 *************************************************************************/
NODE *gedhash_find_xref(HASH_TAB *htab, NODE *dataCtxt)
  {
  NODE *node;
    
  if (node = gedhash_get_rec_xref(htab, ged_get_value(dataCtxt)))
    return(node);
  else
    {  
    ged_error(PTR_ERR, dataCtxt); /* Pointer error.  Could not find pointer. */
    return((NODE *)NULL);
    }
  }

/****************************************************************************
 * NAME: read_tree
 * DESCRIPTION:  Reads in a tree from a GEDCOM file. Returns the pointer to
 *               tree.  If failed, sends an error message and exits.
 *
 * CALLS:        ged_read_tree, GED_ALLOC_POOL
 * CALLED BY:    ged_load_grammar
 ***************************************************************************/
NODE * read_tree(fg, status, limit)
  FILE *fg;
  short *status;
  unsigned long limit;
  {
  NODE *node;
  char *dataBuf;

  if ((!(dataBuf = GED_ALLOC_POOL(limit)))
    || ((node = ged_read_tree(fg, dataBuf, limit, status))==(NODE *)NULL)
    || (*status == REC_TOO_LONG))
    {
    ged_error(304, (NODE *)NULL);
    return((NODE *)NULL);
    }
  return(node);
  } /* End read tree */

/****************************************************************************
 *
 * NAME: gedhash_get_rec_xref
 *
 * DESCRIPTION:  Returns the record pointed to by the xref found in value. For
 * example, given a line, "1 HUSB @I2@".  If you passed the argument, "@I2" into
 * the "value" parameter, gedhash_get_rec_xref() would return a NODE pointer to 
 * "0 @I2@ INDI".  
 *
 * RETURN:  NODE pointer to record, or NULL if no success.
 *
 * PARAMETERS:
 *   htab:  input: Hash table
 *
 *   value: input: Contains the xref of the record that will be returned.
 *          ouput: None.
 ****************************************************************************/
NODE *gedhash_get_rec_xref(HASH_TAB *htab, char *value)
  {
  long seek;
  short status;
  NODE *node = NULL; /* returns this */
  static int flag = TRUE;
  static POOL *pool;
  POOL *oldPool;
  
      if (!htab || !value || !*value)
          return NULL;
      if (flag)
          {
          flag = FALSE; /* don't keep opening pool */
          pool = ged_create_pool((unsigned)GLIMIT, sizeof(NODE));
          }
      oldPool = ged_get_pool();
      ged_set_pool(pool);
      GED_RESET_POOL();
      if (!(((seek = gedh_ptrHash(htab, value)) != -1L) && /* Get lseek of record */
          (fseek(htab->fg, seek, SEEK_SET)==0) && /* seek to record in gedcom file */
          (node = read_tree(htab->fg, &status, GLIMIT)))) /* read record into interal form */
          {
          node = NULL;
          }
      ged_set_pool(oldPool);
      return(node);
      }

/****************************************************************************
 *
 * NAME: gedh_ptrHash
 *
 * DESCRIPTION:  Returns the lseek position of the record containing the xref found in value.
 * 
 * RETURN:  The lseek position of the record, else -1 if unsuccessful
 * 
 * CALLED BY: gedhash_get_rec_xref()
 * 
 * PARAMETERS:
 * 
 * htab:  input: Pointer to HASH_TAB struct.
 *       output: htab->ged_ext may change if gedcom spans multiple disks.
 * 
 * key:   input: Lookup key. Contains xref of the record being searched.
 * 
 ****************************************************************************/
long gedh_ptrHash(HASH_TAB *htab, char *key) /* GEDCOM file */
  {
  LEN valLen;
  KSIZ i = (KSIZ)0; /* must stay type KSIZ!!!! */
  int status; /* status of gedhash_lookup */
  SEEK *lookup_val;
  char val[SEEK_SIZ];
  
  if (*key == '@')
    key++; /* Get past first at sign if there is one */
  while (*(key + (int)i++) != '@'); /* get length of pointer without '@'*/
  lookup_val = (SEEK *)gedhash_lookup(htab, key, i - (KSIZ)1, val, &valLen, &status);
  if (status == SUCCESS)
    {
    gedh_floppy_prompt(htab, lookup_val->ged_ext); /* Check for change in multidisk gedcom file */
    return lookup_val->seek_position;
    }
  return((long)-1);
  }


/************************************************************************
 * NAME: gedh_dCopy
 *
 * DESCRIPTION: Copies key, val, keyLen & valLen into the buffer
 *              that will be written to disk.
 *
 * CALLED BY:   insertDisk
 * 
 * PARAMETERS:
 * 
 * kv:  input: Points to the current location in the hash table block.
 *     output: None.
 * 
 ************************************************************************/
void gedh_dCopy(KVDISK *kv, char *key, KSIZ keyLen, char *val, LEN valLen)
  {
  char *local_buf = (char *)kv;

  kv->keyLen = keyLen;
  kv->valLen = valLen;

  /* insert key in hash table */
  memcpy((char *)(local_buf + (int)sizeof(KVDISK)), key, keyLen);

  /* insert value in hash table (Copy value into end of hash cell) */
  memcpy((char *)(local_buf + (int)sizeof(KVDISK)) + keyLen, val, valLen);
  }

/************************************************************************
 * NAME: gedh_kvGet
 *
 * DESCRIPTION: loads in values from buffer. Loads key & val into
 * the KVDISK struct.
 * 
 * A disk-based hash table is made up of a lot of blocks.
 * One block at a time will be in memory during a function call that processes
 * a disk-based hash table.  Each record in a block includes a KVDISK structure
 * variable.  The fields in this struct contain the key and value lengths (&
 * currently the total record length, not including the length of the KVDISK
 * struct). The 1st part of a record is a KDISK struct, the latter part the key
 * and val fields.
 *
 * CALLED BY:   insertDisk
 *
 * PARAMETERS:
 *
 * kv:  input: Points to the current location in the hash table block.
 *     output: None.
 * 
 * dkey: input: Location to copy key stream to.
 *     output: the key stream from the block is copied here.
 * 
 * dval: input: Location to copy val stream to.
 *     output: The key stream from the block is copied here.
 ************************************************************************/
void gedh_kvGet(KVDISK *kv, char **dkey, char **dval)
  {
  if (kv && (kv->keyLen || kv->valLen) ) /* If we have something to load in */
    {
    char *local_buf = (char *)kv;
  
    *dkey = (char *)(local_buf + (int)sizeof(KVDISK));
    *dval = (char *)(*dkey + (int)kv->keyLen);
    }
  }

/************************************************************************
 * NAME: gedh_diskSetup
 *
 * DESCRIPTION: Initializes a new hash table. Creates hash table on disk,
 * copies from memory, destroys pool.  New size of h table = total number
 * of records.
 *
 * CALLED BY:   gedh_transfer_to_disk, ged_create_table
 *
 *
 * PARAMETERS:
 * 
 * htab:  input: Pointer to hash table struct.
 *       output: Most members of the struct are changed.
 * 
 ************************************************************************/
void gedh_diskSetup(HASH_TAB *htab)
  {
  char fill[BLKLEN]; /* fill the new blocks with zeros */
  register HSIZ i; /* Used in for statement to traverse rows in hash table */
  LEN newh = (LEN)0;      /* new size of hash table on disk */
  static char template[] = "GHXXXXXX"; /* Temporary file name for hash table */

  /* open hash file and fill it with NULLS */
  if (! htab->hfn || ! *(htab->hfn)) /* open temp file if no file name passed */
    {
    strcpy(htab->hfn, template);
    mktemp(htab->hfn);
    htab->fh =ged_eopen(htab->hfn, O_RDWR|O_BINARY|O_CREAT,S_IWRITE|S_IREAD);
    }
  memset(fill,'\0',BLKLEN); /* Not neccesary but nice */
  ged_elseek(htab->fh, 0L, SEEK_SET); /* Position to beginning */
  ged_ewrite(htab->fh, fill, HEADER_SIZ); /* Make space for header */
  /* ged_elseek(htab->fh, 0L + HEADER_SIZ, SEEK_SET); */

  /* Fill up hash table */
  for (i=(HSIZ)0; i < htab->h; i++)
      ged_ewrite(htab->fh, fill, (unsigned int)BLKLEN); /* write the whole block*/

  /* Put identifing record on front of htab & copy onto front of 1st part of record */
  ged_elseek(htab->fh, 0L, SEEK_SET);
  ged_ewrite(htab->fh, "GED_HASH_TAB 1\n", 15); /* Make space for header */
  gedh_save_htab_struct(htab);
  }

/************************************************************************
 * NAME: gedh_transfer_to_disk
 *
 * DESCRIPTION: Creates hash table on disk
 * copies from memory, destroys pool.  New size of h table = total number
 * of records, not to exceed MAX_FILE_H.
 *
 * CALLED BY:   gedhash_insert
 *
 * 
 * PARAMETERS:
 * 
 * htab:  input: Pointer to hash table struct.
 *       output: Most members of the struct are changed.
 * 
 ************************************************************************/
void gedh_transfer_to_disk(HASH_TAB *htab)
  {
  HASH_TAB disk_htab; /* Hash table that is copied onto disk */

  disk_htab = *htab; 

  /* New h value must not be greater than MAX_FILE_H */
  disk_htab.h = (GEDHASH_COUNT(htab) > MAX_FILE_H) ? MAX_FILE_H : GEDHASH_COUNT(htab);
  disk_htab.h = (htab->h > disk_htab.h) ? htab->h : disk_htab.h; /* At least as big as old */

  disk_htab.h = gedh_prime(disk_htab.h); /* Make sure this is a prime number */
  disk_htab.key_count = 0L;  /* There is nothing in this hash table yet */
  gedh_diskSetup(&disk_htab); /* Inititalizes hash table for disk */
  gedhash_copy(&disk_htab, htab); /* Copy memory h table to disk h table */
  *htab = disk_htab; /* Copy disk_htab values into htab */
  }

/************************************************************************
 * 
 * NAME: gedhash_insert
 *
 * DESCRIPTION: Adds a new key/value pair to the hash table.  Will overwrite
 * an existing pair of the same key. If the value of keyLen or valLen is
 * zero, it is assumed that their counterpart, key or val is NULL-terminated.
 * 
 * RETURN VALUE:  Returns zero if unsuccessful, or if a duplicate key is found, 
 * the return value is 2, otherwise 1 is returned.
 * 
 * PARAMETERS:
 * 
 * htab:  input: Pointer to HASH_TAB struct.
 *       output: htab->key_count incremented upon success.  If memory-based,
 *               htab->table gets the key/val pair.  If gedh_diskSetup() is called
 *               then most of the members of this struct is changed.
 *               
 * key:   Input: Look-up key
 *       output: None
 * 
 * keyLen: Input: Length in bytes of key.
 * 
 * val:   Input: None
 *       output: The value associated with key.
 * 
 * valLen: Input: Length in bytes of val.
 *
 ************************************************************************/
int gedhash_insert(HASH_TAB *htab, void *key, KSIZ keyLen, void *val, LEN valLen)
  {
  POOL *oldPool = ged_get_pool(); /* remember what the old pool is */
  int result=0;

  SET_LENGTH(key, keyLen); /* If keyLen == zero, set it to the strlen of key */
  SET_LENGTH(val, valLen);

  if (htab->pool)
      ged_set_pool(htab->pool); /* make current pool the hash pool */
  if (!htab->fh) /* if not hashing to disk yet */
    {
    /* if calling the memory hasher would use too much memory: */
    if ((htab->memoryLeft - (MEMSIZ)(KVSIZE+keyLen+valLen)) < (MEMSIZ)1)
      {  
      gedh_transfer_to_disk(htab); /* copy hash file to disk */
      GED_DESTROY_POOL(); /* now destroy memory hash pool */
      htab->pool = (POOL *) NULL;
      }
    else
      {
      result = gedh_insertMem(htab, (char *)key, keyLen, (char *)val, valLen);
      ged_set_pool(oldPool); /* reset to old pool */
      if (result == SUCCESS)  /* Don't increment key_count if dup. key found */
        (htab->key_count)++; /* Keeps track of number of keys in this hash table */
      return(result);
      }
    }
  result = gedh_insertDisk(htab, (char *)key, keyLen, (char *)val, valLen);
  if (result == SUCCESS)  /* Don't increment key_count if dup. key found */
    (htab->key_count)++; /* Keeps track of number of keys in this hash table */
  ged_set_pool(oldPool); /* reset to old pool */
  return(result);
  }

/************************************************************************
 * NAME: gedh_kvnext
 *
 * DESCRIPTION: A disk-based hash table is made up of a lot of blocks.
 * One block at a time will be in memory during a function call that processes
 * a disk-based hash table.  Each record in a block includes a KVDISK structure
 * variable.  The fields in this struct contain the key and value lengths (&
 * currently the total record length, not including the length of the KVDISK
 * struct). The 1st part of a record is a KDISK struct, the latter part the key
 * and val fields.
 * 
 * This function returns a pointer to the next record in a block.
 * 
 * RETURN:  As stated above, this function returns a pointer to the next record
 *          in a block.
 * 
 * CALLED BY:   insertDisk, gedh_diskSetup
 * 
 *   PARAMETERS:
 *   kv:  input: Starting location of record in disk Buffer.
 *       output: Location of next record in Buffer.
 * 
 ************************************************************************/
KVDISK *gedh_kvnext(KVDISK *kv)
  {
  char *local_buf = (char *)kv;
  /* Used to word-align on cyber */
  char *align = (char *)(local_buf + (int)sizeof(KVDISK) + (int)(kv->totalLen));
  #ifdef CYBER  /* word align for CYBER */
  long mod = (unsigned long)align % 8;

  return  (KVDISK *)( (mod ) ? align + (8 - mod ) : align);
  #endif

  return (KVDISK *)align;
  }

/************************************************************************
 * NAME: gedh_kvCopy
 *
 * DESCRIPTION: Points kv->key & kv->val to right location then copies
 *              key, val, keyLen, valLen to the kv struct
 *
 * CALLED BY:   gedh_insertMem, gedh_insertDisk
 *
 * PARAMETERS:
 *
 * kv:  input: Points to the current location in the hash table block.
 *     output: None.
 * 
 * key: input: buffer which is copied into kv->key
 *     output: None
 * 
 * keyLen: input: Length in bytes of key
 *     output: None
 * 
 * val: input: buffer which is copied into kv->key
 *     output: None
 * 
 * valLen: input: Length in bytes of val
 *     output: None
 *
 ************************************************************************/
void gedh_kvCopy(KVPAIR *kv, char *key, KSIZ keyLen, char *val, LEN valLen)
  {
  kv->key = ((char *)kv + (int)sizeof(KVPAIR)); /* Point kv->key to its proper address */
  kv->val = (char *)(kv->key + keyLen); /*Point kv->val to its proper address*/
  memcpy(kv->val, val,valLen); /* insert value in hash table */
  kv->valLen = valLen;
  memcpy(kv->key, key, keyLen); /* insert key in hash table */
  kv->keyLen = keyLen;
  }

/************************************************************************
 * NAME: gedh_insertMem
 *
 * DESCRIPTION: Inserts a new key/value into memory-based hash table
 *
 * CALLED BY:   gedhash_insert()
 * 
 * RETURN VALUE:  Returns zero if unsuccessful, or if a duplicate key is found, 
 * the return value is 2, otherwise 1 is returned.
 * 
 * 
 * PARAMETERS:
 * 
 * htab:  input: Pointer to HASH_TAB struct.
 *       output: htab->table gets the key/val pair.  If gedh_diskSetup() is called
 *               then most of the members of this struct is changed.
 *               
 * key:   Input: Look-up key
 *       output: None
 * 
 * keyLen: Input: Length in bytes of key.
 *
 * val:   Input: None
 *       output: The value associated with key.
 * 
 * valLen: Input: Length in bytes of val.
 *
 ************************************************************************/
int gedh_insertMem(HASH_TAB *htab, char *key, KSIZ keyLen, char *val, LEN valLen)
  {
  KVPAIR *kv;

  /* look for key */
  for (kv = htab->table[(htab->it_pos = gedh_hash(key, keyLen, htab->h))]; kv; kv = kv->next)
    {
    if ((keyLen==kv->keyLen) && !memcmp(key,kv->key,(size_t)keyLen))  /* found */
      {
      /* key found.  See if space there is big enough to insert key & val.*/
      if (kv->valLen < valLen)
        kv->val = ALLOCMEM(htab,valLen);

      /* Copy value & valLen into hash table  */
      memcpy(kv->val, val, valLen); /* insert value in hash table */
      kv->valLen = valLen;
      return DUP_KEY_FOUND; /* Duplicate key was found */
      }
    }

   /* Key not found in hash table; allocate space to insert. */
  kv = (KVPAIR *)ALLOCMEM(htab, (KVSIZE + keyLen+valLen));
  kv->next = htab->table[ htab->it_pos ];
  htab->table[ htab->it_pos ] = kv;
  gedh_kvCopy(kv, key, keyLen, val, valLen); /* copy values into kv */
  return(SUCCESS); /* success */
  }

/************************************************************************
 * NAME: gedhash_lookup
 *
 * DESCRIPTION: Finds key in the table and returns the associated value (and
 *              length) via the val and valLen reference parameters. 
 *
 * RETURN: Returns pointer to val, else NULL.  If successful, status is set to
 *         SUCCESS=1, else KEY_NOT_FOUND = -1 or HTAB_NOT_FOUND = -2.
 *
 * PARAMETERS:
 * 
 * htab:  input: Pointer to HASH_TAB struct.
 *       output: None
 *               
 * key:   Input: Look-up key
 *       output: None
 * 
 * keyLen: Input: Length in bytes of key.
 * 
 * val:   Input: None
 *       output: The value associated with key.
 * 
 * keyLen: Input: Length in bytes of val.
 *
 * status: Input: None
 *         Output: If successful, returns SUCCESS, else KEY_NOT_FOUND
 *
 ************************************************************************/
void * gedhash_lookup(HASH_TAB *htab, void *key, KSIZ keyLen, void *val, LEN *valLen, int *status)
  {
  SET_LENGTH(key, keyLen); /* If keyLen == zero, set it to the strlen of key */

  /* look for key */
  *status = NO_SUCCESS; /* Pessemistic default */
  if (!htab)
    {  
    *status = HTAB_NOT_FOUND;
    return NULL;
    }
  if (!htab->fh) /* If memory-based hash table */
    {
    KVPAIR *kv;
    for (kv = htab->table[gedh_hash((char *)key, keyLen, htab->h)]; kv; kv = kv->next)
      {  
      if ((keyLen==kv->keyLen) && memcmp((char *)key,kv->key,(size_t)keyLen)==0)  /* found */
        {
        *valLen = kv->valLen;
        memcpy((char *)val, kv->val, kv->valLen);
        *status = SUCCESS;
        return val;
        }
      }
    *status = KEY_NOT_FOUND;
    return NULL;
    }
  return gedh_dlookup(htab, (char *)key, keyLen, val, valLen, status);
  }

/**********************************************************************
 * NAME: gedhash_delete
 *
 * DESCRIPTION:  Finds the key in the hash table and deletes its cell.
 *               Returns zero if successful, non-zero otherwise.
 * 
 * RETURN:  Zero if key not found, else SUCCESS.
 *
 * PARAMETERS:
 *
 * htab:  input: Pointer to HASH_TAB struct.
 *       output: htab->key_count decremented.  If memory based, htab->table is
 *               changed. 
 * 
 * key:   Input: If this key is found, record will be deleted from hash table
 *       output: None
 * 
 * keyLen: Input: Length in bytes of key.
 *
 **********************************************************************/
int gedhash_delete(HASH_TAB *htab, void * key, KSIZ keyLen)
  {
  /* look for key */
  int return_value;

  SET_LENGTH(key, keyLen); /* If keyLen == zero, set it to the strlen of key */
  if (!htab->fh)  /* If in memory */
    {
    KVPAIR *kv;       /* Current kv record being processed */
    KVPAIR *top;      /* Root node of a collision list */
    KVPAIR *previous; /* Parent node of kv */

    for (kv = top = htab->table[htab->it_pos = gedh_hash(key, keyLen, htab->h)]; kv; kv = kv->next)
      {        
      if (memcmp((char *)key,kv->key,(size_t)keyLen)==0)  /* found */
        {
        if (kv = top) /* If we are @ beginning of collision list */  
          htab->table[htab->it_pos] = kv->next; /* deleted 1st cell */
        else
          previous->next = kv->next; /* Delete kv */
        (htab->key_count)--;  /* Keeps track of number of keys in hash table */
        return_value = SUCCESS; /* successful */
        break;
        }
      previous = kv; /* keep track of the previous kv so I can use it to link.*/
      }
    return_value = NO_SUCCESS; /* Non successful. Did not find match */
    }
  else /* On disk; not in memory */
    return_value = gedh_delete_disk(htab, (char *)key, keyLen);  
  if (return_value == SUCCESS)
    (htab->key_count)--;
  return return_value;
  }

/****************************************************************************
 * NAME: gedh_hash
 *
 * DESCRIPTION: hashes the key and outputs it to a certain row in the hash table
 * 
 * RETURN:  The hashed value (location in the hash table) of key
 * 
 * PARAMETERS:
 * 
 * key:   Input: Look-up key being hashed.
 *       output: None
 * 
 * keyLen: Input: Length in bytes of key.
 * 
 * h:   Input: Size of hash table, which is number of rows in hash table; number of
 *             beginnings of collision or potential collision lists.
 *       output: The value associated with key.
 *
 ****************************************************************************/
HSIZ gedh_hash(char *key, KSIZ keyLen, HSIZ h)
  {
  register KSIZ ii = (KSIZ)0,
           i;

  for (i = (KSIZ)0; i < keyLen; i++)
    ii = ii << 1^*key++;
  return( ( (ii < (KSIZ)0) ? -ii : ii ) % h ); 
  }

/****************************************************************************
 *
 * NAME: gedhash_destroy_table
 * 
 * Description:  Destroys the table, releases all memory used by it, and deletes
 *               the hash file on disk, if any.  
 * 
 * Return Value: Returns 1 if successful, otherwise zero.
 * 
 * PARAMETERS:
 *
 * htab:  input: Pointer to HASH_TAB struct.
 *       output: Destroys the table, releases all memory used by it, and deletes
 *               the hash file on disk, if any.
 * 
 ****************************************************************************/
int gedhash_destroy_table(HASH_TAB *htab)
  {
  POOL *pool;  /* current pool */

  if (!htab)
    return NO_SUCCESS;
  if (!htab->fh) /* if using memory-based table */
    {
    if (htab->pool)
        {
        pool = ged_get_pool();
        ged_set_pool(htab->pool); /* set current pool to the htab */
        GED_DESTROY_POOL();
        htab->pool = (POOL *) NULL;
        ged_set_pool(pool); /* set current pool back to what it was before */
        }
    }
  else
    {
    if (gedhash_close_table(htab) != 0) /* Not a success */
      return NO_SUCCESS;  /* Could not close & therefore delete hash file */
    if (!htab->isMemory) /*If user passed in a file name to gedhash_create_tabl*/
      if (remove(htab->hfn))
        {
        ged_error(REMOVE_ERR, (NODE *)NULL); /* Could not delete permanent hash file. */
        return NO_SUCCESS;
        }  
    }
  return SUCCESS; 
  }

/************************************************************************
 *
 * NAME: gedhash_close_table
 * 
 * Description:  Closes a hash table file created with a permanent name. If
 * this function is called it is assumed that there is a hash
 * table file to close.
 * 
 * htab->hfn is deleted if the hash table was opened in memory-based mode, and then
 * was copied onto a temporary file when htab->maxMemory was reached.
 * 
 * Return Value: Returns zero upon success.  Returns -1 if no hash table
 *               file exits.  Returns 1 if file cannot be deleted or closed.
 * 
 * PARAMETERS:
 *
 * htab:  input: Pointer to HASH_TAB struct.
 *       output: htab->fh closed. htab->hfn deleted if temporary.
 *
 ************************************************************************/
int gedhash_close_table(HASH_TAB *htab)
  {
  if (htab && htab->fh) /* if using disk-based table */
    {  
    if (gedh_save_htab_struct(htab) != SUCCESS) /* Write header to permanent file */
      ged_error(WRITE_ERR,(NODE *)NULL);
      
    ged_eclose(htab->fh); /* Close file-based hash table */
    if (htab->isMemory) /* If we should delete a temporary table file */
      if (remove(htab->hfn) == -1)
        {
        ged_error(REMOVE_ERR, (NODE *)NULL); /* Could not delete temporary hash file. */
        return 1; /* file cannot be deleted or closed. */
        }  
    return 0;  /* Success */
    }
  else
    return -1; /* Hash table was not on disk */
  }

/************************************************************************
 *
 * NAME: gedh_prime
 * 
 * Description: This checks to see if the argument, h, is prime.  If not
 * it returns the next lower prime number.
 * 
 * 
 * Return Value: Returns h if it is a prime number, else the next lowest
 * numbwr that is a prime number.
 *               
 * 
 * PARAMETERS:
 *
 * h:  input:   Size of hash table
 *
 ************************************************************************/
HSIZ gedh_prime(HSIZ h)
  {
  HSIZ i, possible_divider;
  int flag = 0;
  
  for (i = h; i; i--)
    {  
    for (possible_divider = 2; possible_divider < i; possible_divider++)
      {
      /* if i is not a prime (or if i is divided evenly by possible_divider) */
      if (!(i % possible_divider )) 
        {
        flag = 1; /* Not a prime */
        break;    
        }
      }
    if (!flag) /* If i is a prime */
      return i;
    flag = 0;
    }
  return (HSIZ)1;  /* At least 1 is a prime! */
  }


/************************************************************************
 *
 * NAME: filelength
 * 
 * Description: Returns the length of file associated with handle. Used
 * for CYBER becasue CYBER has no filelength().
 * 
 * Returns:   the file length in bytes, or -1L if an error occurs.
 *            errno: EBADF
 * 
 * PARAMETERS:
 *
 * h:  input:   Size of hash table
 *
 ************************************************************************/
#ifdef NON_ANSI
long filelength( int handle )
  {
  long original_pos = lseek(handle, 0L, SEEK_CUR); /* Get current position in file */
  long file_len; /* Length of file.  This is returned */

  if ( (original_pos == -1L) 
      || ((file_len = lseek(handle, 0L,  SEEK_END) ) == -1L) /* get file size */
      || (lseek(handle, original_pos,  SEEK_SET)  == -1L) ) /* Seek to orig pos */
    {  
    ged_error(SEEK_ERR, (NODE *)NULL); /* Seek error */
    return -1L;
    }
  return file_len;
  }

#endif


/************************************************************************
 * NAME: gedh_word_align
 *
 * DESCRIPTION: Aligns an address.  This is for the CYBER
 *
 *
 * RETURN:  Pointer to right address
 *
 *
 *   PARAMETERS:
 *
 ************************************************************************/
#ifdef CYBER
void *gedh_word_align(void *address)
  {
  long mod = (long)address % 8;

  return  (void *)( (mod ) ? address + (8 - mod ) : address);
  }
#endif
