/***********************************************************************
*$Header:   J:/aim/gedcom/gedlib/vcs/gedrdmem.c_V   1.6   25 Jan 1996 08:23:24   ODJENSEN  $
*
*$config$="/K! /L/* /R* /Mgedrdmem.c"
*
*   FILE NAME: GEDRDMEM.C
*
*   DESCRIPTION:
*       This file contains functions for reading from files.
*
*   ROUTINES:
*
*
*$Log:   J:/aim/gedcom/gedlib/vcs/gedrdmem.c_V  $
 * 
 *    Rev 1.6   25 Jan 1996 08:23:24   ODJENSEN
 * Added function headers, now only reports each type of error once per line f
 * most errors.
 * 
 *    Rev 1.5   22 Jan 1996 17:03:18   ODJENSEN
 * Fixed problem of reporting multiple errors per line, problem with reading
 * 0 TRLR record, and now allows tabs in value.
 * 
 *    Rev 1.4   22 Jan 1996 11:28:28   ODJENSEN
 * 
 * Updated ged_read_nodetree to be more compliant with the new gedcom standard
 * and cleaned up the code.
 * 
 *    Rev 1.3   18 Jan 1996 08:03:56   ODJENSEN
 * Changed to new-style declarators, cleaned up a couple of warnings.
 * 
 *    Rev 1.2   07 Jun 1994 08:09:32   odjensen
 * Changed function to use fgetc instead of fscanf.
 * 
 *    Rev 1.1   06 Nov 1992 10:43:48   fhdodj
 * Fixed header information.
 * 
 *    Rev 1.0   05 Aug 1992 19:28:32   fhdodj
 * Initial revision.
 * 
 *    Rev 1.0   28 Jul 1992 09:34:36   fhdodj
 * Initial revision.
 * 
 *    Rev 1.12   03 Mar 1992 09:44:54   fhdodj
 * Replaced fpos with -bytesRead in NON_ANSII part of ged_read_line.
 * 
 *    Rev 1.11   03 Mar 1992 09:35:14   fhdodj
 * Removed fflush and ftell from file.
 * 
 *    Rev 1.10   21 Oct 1991 09:06:44   fhdodj
 * Changed char to byte.
 * 
 *    Rev 1.9   09 Aug 1991 10:24:24   fhdodj
 * Changed condition in ged_read_tree so that LAST_RECORD is now returned.
 * 
 *    Rev 1.8   26 Jul 1991 13:31:26   fhdodj
 * Fixed status return from ged_read_line and ged_read_level.  Didn't return
 * LAST_RECORD correctly.
 * 
 *    Rev 1.7   28 Jun 1991 14:08:10   fhdkrf
 * Added keywords for PolyDoc
 * 
 *    Rev 1.6   25 Jun 1991 09:04:14   fhdodj
 * Fixed functions ged_connect_child and ged_connect_sibling.  They were 
 * attaching nodes in the wrong place when given an argument other than -1.
 * 
 *    Rev 1.5   18 Jun 1991 13:39:56   odj
 * Added check in ged_read_line so we dont get an error for invalid character
 * when there is a carriage return or newline.
 * 
 *    Rev 1.4   18 Jun 1991 11:08:12   odj
 * Changed error number ranges.
 * 
 *    Rev 1.3   17 Jun 1991 16:03:00   odj
 * Added IO status checks.  Added calls to ged_error() function.  Also added
 * checking in ged_read_line function for invalid tags and cross references
 * being read.  Changed arguments to ged_read_level() function so that status is
 * now being checked in this function also.
 * 
 *    Rev 1.2   25 Apr 1991 10:06:18   odj
 * changed ged_fprintf() to output a carriage return linefeed insteadof just a 
 * linefeed.
 * 
 *    Rev 1.1   14 Feb 1991 15:02:42   odj
 * void is not allowed by some non ansi compilers (e.g. the CYBER). so I 
 * changed void to be char for non ansi compilers.
 * 
 *    Rev 1.0   20 Dec 1990 08:59:48   odj
 * Initial revision.
***********************************************************************/
#include "gedcom.h"

/********************************************************************
*
*        NAME: ged_read_a_line
*
*        DESCRIPTION:
*            Get the next line from a gedcom file. status
*            REC_TOO_LONG is given when no more space is
*            available in the input buffer.  LAST_RECORD is
*            given when EOF is encountered.  MORE_RECORDS is
*            given when there are more records available to read.
*            END_OF_FILE is returned if EOF is encountered before
*            anything is read. If the line is too long, the extra
*            characters are thrown away.
*
*        RETURN VALUE:
*            The number of bytes read.
*
********************************************************************/
static long ged_read_a_line(FILE *fp, short *status, byte *line, int size)
{
    long bytesRead = 0;
    int pos = 0;
    int c;
    
    if (size < MAX_LINE_LEN + 1)
        return -1;
    memset(line, '\0', size);
    while (isspace(c = fgetc(fp)))
        bytesRead++;
    if (c != EOF)
        bytesRead++;
    if (feof(fp))
    {
        *status = END_OF_FILE;
        return -1;
    }
    if (ferror(fp) && !ged_error(103, NULL))
        return -1;
    line[pos++] = (byte)c;
    while (((c = fgetc(fp)) != '\n') && c != '\r' && c != EOF &&
           pos <= MAX_LINE_LEN)
    {
        if (ferror(fp) && !ged_error(104, NULL))
            return -1;
        line[pos++] = (byte)c;
        bytesRead++;
    }
    if (c != EOF)
        bytesRead++;
    line[pos] = '\0';
    if (ferror(fp) && !ged_error(105, NULL))
        return -1;
    if (pos > MAX_LINE_LEN && !ged_error(297, NULL))
        return -1;
    else if (pos > MAX_LINE_LEN)
    {
        while ((c = fgetc(fp)) != '\n' && c != '\r' && c != EOF)
            bytesRead++;
        if (c != EOF)
            bytesRead++;
        if (ferror(fp) && !ged_error(106, NULL))
            return -1;
    }
    if (feof(fp) && !ged_error(296, NULL))
        *status = END_OF_FILE;
    while (line[pos] == '\n' || line[pos] == '\r')
        line[pos--] = '\0';
    while (isspace(c = fgetc(fp)))
        bytesRead++;
    if (c != EOF)
        fseek(fp, -1, SEEK_CUR);
    else if (pos)
        *status = LAST_RECORD;
    else
        *status = END_OF_FILE;
    return bytesRead;
}

/********************************************************************
*
*        NAME: ged_get_level
*
*        DESCRIPTION:
*            Get the level number from a line.  return the next line
*            position after the level number, and set level to be
*            the level number read.
*
*        RETURN VALUE:
*            The address of the next part of the line. (crossref ID or
*            tag)
*
********************************************************************/
static byte * ged_get_level(byte *line, int *level)
{
    *level = 0;
    if (isdigit(*line))
    {
        while (isdigit(*line))
            *level = (*level * 10) + *line++ - '0';
        if (*line == ' ')
            line++;
    }
    else
        *level = -1;
    return line;
}

/********************************************************************
*
*        NAME: ged_verify_level
*
*        DESCRIPTION:
*            Get the level number from a line, and check that a valid value
*            was encountered.
*
*        RETURN VALUE:
*            The address of the next part of the line. (crossref ID or
*            tag)
*
********************************************************************/
static byte * ged_verify_level(byte *line, int *level)
{
    line = ged_get_level(line, level);
    if (*level == -1 &&!ged_error(280, NULL)) /* no level number */
        return 0;
    if (*level == -1)
        *level = 0;
    return line;
}

/********************************************************************
*
*        NAME: ged_verify_xref
*
*        DESCRIPTION:
*            Check that the xref if it exists contains only valid
*            characters, is a correct length, and is terminated
*            properly.
*
*        RETURN VALUE:
*            The address of the next part of the line. (The tag)
*
********************************************************************/
static byte * ged_verify_xref(byte *line, int errorBase)
{
    int xrefLen = 0;
    int badCharFlag = 0;

    if (*line == '@')
    {
        line++;
        xrefLen++;
        while (*line && *line != '@')
        {
            if (*line < ' ' && *line != 255)
                badCharFlag = 1;
            xrefLen++;
            line++;
        }
        if (badCharFlag && !ged_error(errorBase + 1, NULL))
            return 0;    /* invalid character in xref ID */
        if (*line != '@' && !ged_error(errorBase + 5, NULL))
            return 0;    /* no terminating @ */
        line++;
        xrefLen++;
        if ((xrefLen > MAX_CROSSREF_LEN || xrefLen < MIN_CROSSREF_LEN) &&
            !ged_error(errorBase + 2, NULL))
        {
            return 0;    /* xref incorrect size */
        }
        if (*line == ' ')
            line++;
    }
    return line;
}

/********************************************************************
*
*        NAME: ged_verify_tag
*
*        DESCRIPTION:
*            Check that the tag exists and contains only valid
*            characters, and is a correct length.
*
*        RETURN VALUE:
*            The address of the next part of the line. (The value)
*
********************************************************************/
static byte * ged_verify_tag(byte *line)
{
    int tagLen = 0;
    int badCharFlag = 0;
    
    while (*line != ' ' && *line)
    {
        if (!isalnum(*line) && *line != '_' && *line != '+' && *line != '-')
            badCharFlag = 1;
        tagLen++;
        line++;
    }
    if (badCharFlag && !ged_error(203, NULL))
        return 0;    /* invalid character in tag */
    if ((tagLen > MAX_TAG_LEN || tagLen < MIN_TAG_LEN) &&
        !ged_error(204, NULL))
    {
        return 0;    /* incorrect tag length */
    }
    if (*line == ' ')
        line++;
    return line;
}

/********************************************************************
*
*        NAME: ged_verify_escape
*
*        DESCRIPTION:
*            If there is an escape sequence, check that the it
*            contains only valid characters, is a correct length,
*            and is terminated properly. 
*
*        RETURN VALUE:
*            The address of the next part of the line.
*
********************************************************************/
static byte * ged_verify_escape(byte *line)
{
    int escapeLen = 0;
    int badCharFlag = 0;

    if (*line == '@' && *(line + 1) == '#')
    {
        line += 2;
        escapeLen += 2;
        while (*line && *line != '@')
        {
            if (!isalnum(*line) && *line != '_' && *line != ' ')
                badCharFlag = 1;
            escapeLen++;
            line++;
        }
        if (badCharFlag && !ged_error(295, NULL))
            return 0;    /* invalid char in escape sequence */
        if (*line != '@' && !ged_error(293, NULL))
            return 0;    /* no terminating @ in escape sequenct */
        if (escapeLen > MAX_ESCAPE_LEN && !ged_error(294, NULL))
            return 0;    /* escape sequence too long */
        line++;
        if (*line == ' ')
            line++;
    }
    return line;
}

/********************************************************************
*
*        NAME: ged_verify_value
*
*        DESCRIPTION:
*            If there is a value, check that if there is a pointer
*            value, that it is the only value on the line, the value
*            contains only valid characters, is a correct length,
*            and verify any escape sequences that appear.
*
*        RETURN VALUE:
*            0 if error, 1 if OK.
*
********************************************************************/
static int ged_verify_value(byte *line)
{
    int badCharFlag = 0;

    if (*line == '@' && *(line + 1) != '#')
    {
        if ((line = ged_verify_xref(line, 220)) == 0)
            return 0;  /* invalid pointer on line. */
        if (*line && !ged_error(292, NULL))
            return 0;  /* pointer only is allowed in value */
    }
    while (*line)
    {
        if ((*line < ' ' && *line != 0x09) || *line == 0xFF)
            badCharFlag = 1;
        if ((*line == '@') && (*(line + 1) == '@'))
            line += 2;
        else if ((*line == '@') && (*(line + 1) == '#') &&
                 ((line = ged_verify_escape(line)) == 0))
        {
            return 0;   /* invalid escape sequence */
        }
        else if (*line == '@' && !ged_error(290, NULL))
            return 0;   /* pointer value must be only item in value */
        else
            line++;
    }
    if (badCharFlag && !ged_error(291, NULL))
        return 0;   /* invalid character in value */
    return 1;
}

/********************************************************************
*
*        NAME: ged_verify_line
*
*        DESCRIPTION:
*            Verify that each piece of the line conforms to the gedcom
*            standard chapter 1 specification.
*
*        RETURN VALUE:
*            The address of the first part of the line after the level.
*
********************************************************************/
byte * ged_verify_line(byte *line)
{
    int level = -1;
    int tagLen  = 0;
    byte *lineStart = 0;

    if (!line || !*line)
        return 0;
    if (strlen(line) > MAX_LINE_LEN && !ged_error(297, NULL))
        return 0;    /* line too long */
    if ((lineStart = line = ged_verify_level(line, &level)) == 0)
        return 0;    /* invalid level */
    if ((line = ged_verify_xref(line, 200)) == 0)
        return 0;    /* invalid xref */
    if ((line = ged_verify_tag(line)) == 0)
        return 0;    /* invalid tag */
    if (ged_verify_value(line) == 0)
        return 0;    /* invalid value */
    return lineStart;
}

/********************************************************************
*
*        NAME: ged_read_node
*
*        DESCRIPTION:
*            read a line from the file, set status, verify that the 
*            line conforms with the gedcom standard, and create a node.
*
*        RETURN VALUE:
*            The level of the line read.
*
********************************************************************/
static int ged_read_node(FILE *fp, short startLevel, short *status, NODE **node)
{
    int   level = 0;
    byte  buf[300];
    long  bytesRead;
    byte *lineStart = buf;

    *node = 0;
    *status = MORE_RECORDS;
    memset(buf, '\0', 300);
    bytesRead = ged_read_a_line(fp, status, buf, 300);
    if (bytesRead <= 0)
        return -10;
    ged_get_level(buf, &level);
    if (level < 0)
        level = 0;
    if (level <= startLevel)
        {
        fseek(fp, -bytesRead, SEEK_CUR);
        *status = MORE_RECORDS;
        return level;
        }
    lineStart = ged_verify_line(buf);
    if (lineStart && feof(fp))
        *status = LAST_RECORD;
    else if (!lineStart && feof(fp))
        *status = END_OF_FILE;
    if (lineStart)
        *node = ged_copy_node(lineStart, NULL, NULL);
    return level;
}

/********************************************************************
*
*        NAME: ged_read_node_tree
*
*        DESCRIPTION:
*            Create a node corresponding to each line read from the file,
*            attach the node in the tree where it belongs, and set status.
*
*        RETURN VALUE:
*            The root node of the tree read in.
*
********************************************************************/
NODE * ged_read_nodetree(FILE *fp, short *status)
{
    int   level, prev;
    int   startLev = 0;
    NODE *root;
    NODE *node;
    NODE *tmpNode;
    int x;

        if (!fp)
            return(NULL);
        prev = startLev = ged_read_node(fp, -1, status, &node);
        root = tmpNode = node;
        while ((*status != LAST_RECORD) && (*status != END_OF_FILE) &&
               ((level = ged_read_node(fp, (short)startLev, status,
                                       &node)) > startLev))
            {
            if ((prev < level) && ((level - prev) != 1))
                {    /* not next child of prev      */
                *status = MISSING_LEVEL;
                if (!ged_error(281, node))
                    return root;    /* missing level number */
                }
            else
                {
                if (prev < level)
                    tmpNode = ged_connect_child(tmpNode, node, -1);
                else
            {
            for (x = prev - level; x; x--)
                        tmpNode = ged_get_parent(tmpNode);
                    tmpNode = ged_connect_sibling(tmpNode, node, -1);
                    }
                prev = level;
                }
            }
        return(root);
}
