/***********************************************************************
*$Header:   J:/aim/gedcom/gedlib/vcs/gedwrite.c_V   1.4   16 Jan 1996 09:04:34   ODJENSEN  $
*
*$config$="/K! /L/* /R* /Mgedwrite.c"
*
*   FILE NAME: GEDWRITE.C
*
*   DESCRIPTION:
*       This file contains functions for writing to files.
*
*   ROUTINES:
*
*
*$Log:   J:/aim/gedcom/gedlib/vcs/gedwrite.c_V  $
 * 
 *    Rev 1.4   16 Jan 1996 09:04:34   ODJENSEN
 * Cleaned up warnings, and changed to new-style declarators.
 * 
 *    Rev 1.3   15 Jun 1994 16:27:46   odjensen
 * changed in ged_sprintf_line *limit-- to be *limit -= 1;
 * 
 *    Rev 1.2   26 Apr 1993 14:57:20   fhdodj
 * Added function ged_buffered to tree.  This function is a little faster
 * than ged_tree_to_file.
 * 
 *    Rev 1.1   30 Nov 1992 09:16:48   fhdodj
 * Fixed problem in ged_sprintf_line where limit was not always correctly
 * updated.
 * 
 *    Rev 1.0   28 Jul 1992 09:33:02   fhdodj
 * Initial revision.
 * 
 *    Rev 1.12   03 Mar 1992 09:44:54   fhdodj
 * Replaced fpos with -bytesRead in NON_ANSII part of ged_read_line.
 * 
 *    Rev 1.11   03 Mar 1992 09:35:14   fhdodj
 * Removed fflush and ftell from file.
 * 
 *    Rev 1.10   21 Oct 1991 09:06:44   fhdodj
 * Changed char to byte.
 * 
 *    Rev 1.9   09 Aug 1991 10:24:24   fhdodj
 * Changed condition in ged_read_tree so that LAST_RECORD is now returned.
 * 
 *    Rev 1.8   26 Jul 1991 13:31:26   fhdodj
 * Fixed status return from ged_read_line and ged_read_level.  Didn't return
 * LAST_RECORD correctly.
 * 
 *    Rev 1.7   28 Jun 1991 14:08:10   fhdkrf
 * Added keywords for PolyDoc
 * 
 *    Rev 1.6   25 Jun 1991 09:04:14   fhdodj
 * Fixed functions ged_connect_child and ged_connect_sibling.  They were 
 * attaching nodes in the wrong place when given an argument other than -1.
 * 
 *    Rev 1.5   18 Jun 1991 13:39:56   odj
 * Added check in ged_read_line so we dont get an error for invalid character
 * when there is a carriage return or newline.
 * 
 *    Rev 1.4   18 Jun 1991 11:08:12   odj
 * Changed error number ranges.
 * 
 *    Rev 1.3   17 Jun 1991 16:03:00   odj
 * Added IO status checks.  Added calls to ged_error() function.  Also added
 * checking in ged_read_line function for invalid tags and cross references
 * being read.  Changed arguments to ged_read_level() function so that status is
 * now being checked in this function also.
 * 
 *    Rev 1.2   25 Apr 1991 10:06:18   odj
 * changed ged_fprintf() to output a carriage return linefeed insteadof just a 
 * linefeed.
 * 
 *    Rev 1.1   14 Feb 1991 15:02:42   odj
 * void is not allowed by some non ansi compilers (e.g. the CYBER). so I 
 * changed void to be char for non ansi compilers.
 * 
 *    Rev 1.0   20 Dec 1990 08:59:48   odj
 * Initial revision.
***********************************************************************/
#include "gedcom.h"

/********************************************************************
*
*        NAME: ged_fprintf_line
*
*        DESCRIPTION:
*            Writes a line to the file pointed to by fp.
*
*        RETURN VALUE:
*            None.
*
*        CALLED BY: ged_tree_to_file
*
********************************************************************/
int ged_fprintf_line(int level, NODE *node, void *tfp, void *unreferencedArg2)
    {
    FILE *fp = (FILE *) tfp;

        if (node->line)
            fprintf(fp, "%d %s\r\n", level, node->line);
        if (ferror(fp))
            ged_error(109, NULL);
        return(1);
    }

/********************************************************************
*
*        NAME: ged_sprintf_line
*
*        DESCRIPTION:
*            Writes a line to gedBuf
*
*        RETURN VALUE:
*            None.
*
********************************************************************/
int ged_sprintf_line(int level, NODE *node, void *tempGedBuf, void *tempLimit)
    {
    byte **gedBuf = (byte **)tempGedBuf;
    byte *ptr = *gedBuf;
    unsigned *limit = tempLimit;

        if (node->line && ((*limit - (node->length + 5)) > 0))
	    {                                    /* ^ level len plus space */
	    sprintf(ptr, "%d %s\n", level, node->line);
	    while (*ptr != '\0')
		{
		*limit -= 1;
		ptr++;
		}
	    ptr++;
            *gedBuf = ptr;
            }
        return(1);
    }

/********************************************************************
*
*        NAME: ged_tree_to_file
*
*        DESCRIPTION:
*            Writes the input tree to the file pointed to by fp.
*
*        RETURN VALUE:
*            None.
*
*        CALLS: ged_apply_tree, ged_fprintf_line
*
********************************************************************/
void ged_tree_to_file(NODE *tree, FILE *fp, unsigned level)
    {

        ged_apply_tree(tree, level, 0, ged_fprintf_line, fp, NULL);
    }

/********************************************************************
*
*        NAME: ged_tree_to_ged
*
*        DESCRIPTION:
*            Writes the input tree to buffer gedBuf. limit is the size
*            of gedBuf.
*
*        RETURN VALUE:
*            Returns the address to the position in the buffer where
*            we finished writing.
*
*        CALLS: ged_apply_tree, ged_sprintf_line
*
********************************************************************/
byte * ged_tree_to_ged(NODE *tree, byte *gedBuf, unsigned limit,
                       unsigned level)
    {
        ged_apply_tree(tree, level, 0, ged_sprintf_line, &gedBuf, &limit);
        return(gedBuf);
    }

/********************************************************************
*
*        NAME: ged_buffered_to_file
*
*        DESCRIPTION:
*            Writes the input tree to the file pointed to by fp.
*
*        RETURN VALUE:
*            None.
*
*        CALLS: ged_get_tree_size, ged_get_next_level, ged_alloc_pool
*
********************************************************************/
void ged_buffered_to_file(NODE *tree, FILE *fp, int level)
    {
    NODE *node = tree;
    int start_level = level;
    int size = 0;
    int offset = 0;
    byte *buffer;
    byte level_num[20];

	size = ged_get_tree_size(tree);
	buffer = ged_alloc_pool(size);
	node = tree;
	level = start_level;
	while(node)
	    {
	    if (level < 10)
		buffer[offset++] = (byte)(level + '0');
	    else
		{
		itoa(level, level_num, 10);
		memcpy(&buffer[offset], level_num, (size=strlen(level_num)));
		offset += size;
		}
	    buffer[offset] = ' ';
	    offset++;
	    memcpy(&buffer[offset], node->line, node->length);
	    offset += node->length;
	    memcpy(&buffer[offset], "\r\n", 2);
	    offset += 2;
	    node = ged_get_next_level(node, &level);
	    }
	fwrite(buffer, 1, offset, fp);
    }
